from __future__ import annotations

import abc
import dataclasses
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Callable
    from .token import OAuthToken


class Eventable(abc.ABC):
    """Abstract for classes that handle events"""

    def __init__(self) -> None:
        self._listeners: list[Callable] = []

    @abc.abstractmethod
    async def _process_event(self, event: BaseEvent) -> None:
        ...


@dataclasses.dataclass
class BaseEvent(abc.ABC):
    """Abstract for event classes"""


@dataclasses.dataclass
class ClientAddEvent(BaseEvent):
    client_id: int
    """0 if app client"""
    client: Eventable
    """The client that was added"""


@dataclasses.dataclass
class ClientUpdateEvent(BaseEvent):
    client: Eventable
    old_token: OAuthToken
    new_token: OAuthToken



"""
This module handles multiple API v2 Client sessions.
"""
from __future__ import annotations

import functools
from typing import TYPE_CHECKING

from . import Client
from ..classes import OAuthToken
from ..classes.events import ClientAddEvent
from ..classes.events import ClientUpdateEvent
from ..classes.events import Eventable

if TYPE_CHECKING:
    from types import TracebackType
    from typing import Any
    from typing import Callable
    from typing import Optional
    from typing import Type
    from typing import Union
    from ..classes.events import BaseEvent


class ClientStorage(Eventable):
    r"""OAuth sessions manager.

    :param \**kwargs:
        See below

    :Keyword Arguments:
        * *client_secret* (``str``)
        * *client_id* (``int``)
        * *base_url* (``str``) --
            Optional, base API URL, defaults to \"https://osu.ppy.sh/api/v2/\"
        * *create_app_client* (``bool``) --
            Optional, whether to automatically create guest clients, defaults to False
        * *app_client* (``aiosu.v2.client.Client``) --
            Optional, guest client, defaults to None
    """

    def __init__(self, **kwargs: Any) -> None:
        super().__init__()
        self.client_secret: str = kwargs.pop("client_secret", None)
        self.client_id: int = kwargs.pop("client_id", None)
        self.base_url: str = kwargs.pop("base_url", "https://osu.ppy.sh/api/v2")
        self.__create_app_client: bool = kwargs.pop("create_app_client", False)
        self.clients: dict[int, Client] = {}
        if _app_client := kwargs.pop("app_client", None):
            self.clients[0] = _app_client
        self._redirected_listeners: list[Callable] = []

    async def __aenter__(self) -> ClientStorage:
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        await self.close()

    def on_client_add(self, func: Callable) -> Callable:
        """
        A decorator that is called whenever a client is added, to be used as:

            @client_storage.on_client_add

            async def func(event: ClientAddEvent)
        """
        self._listeners.append(func)

        @functools.wraps(func)
        async def _on_client_add(*args: Any, **kwargs: Any) -> Any:
            return await func(*args, **kwargs)

        return _on_client_add

    def on_client_update(self, func: Callable) -> Callable:
        """
        A decorator that is called whenever any stored client is updated, to be used as:

            @client_storage.on_client_update

            async def func(event: ClientUpdateEvent)
        """
        self._redirected_listeners.append(func)

        @functools.wraps(func)
        async def _on_client_update(*args: Any, **kwargs: Any) -> Any:
            return await func(*args, **kwargs)

        return _on_client_update

    async def _process_event(self, event: BaseEvent) -> None:
        if isinstance(event, ClientAddEvent):
            for func in self._listeners:
                await func(event)
            return
        if isinstance(event, ClientUpdateEvent):
            for func in self._redirected_listeners:
                await func(event)
            return
        raise NotImplementedError(f"{event!r}")

    @property
    async def app_client(self) -> Client:
        r"""Client credentials guest client.

        :raises NotImplementedError: If no app client is provided and creation is disabled
        :return: Client credentials guest client session
        :rtype: aiosu.v2.client.Client
        """
        if not (_app_client := self.clients.get(0)):
            raise NotImplementedError("Client credential grant creation is still WIP")
            await self._process_event(
                ClientAddEvent(client_id=0, client=self._app_client),
            )

        return _app_client

    def _get_client_args(self) -> dict[str, Union[str, int]]:
        return {
            "client_secret": self.client_secret,
            "client_id": self.client_id,
            "base_url": self.base_url,
        }

    def client_exists(self, client_uid: int) -> bool:
        r"""Checks if a client exists.

        :param client_uid: The owner user ID of the client
        :type client_uid: int
        :return: Whether the client with the given ID exists
        :rtype: bool
        """
        return client_uid in self.clients

    async def add_client(self, token: OAuthToken) -> Client:
        r"""Adds a client to storage.

        :param token: Token object for the client
        :type token: aiosu.classes.token.OAuthToken
        :return: The added client
        :rtype: aiosu.v2.client.Client
        """
        client = Client(token=token, **self._get_client_args())
        client._listeners.append(self._process_event)
        client_user = await client.get_me()
        self.clients[client_user.id] = client
        await self._process_event(
            ClientAddEvent(client_id=client_user.id, client=client),
        )
        return client

    async def get_client(self, **kwargs: Any) -> Client:
        r"""Gets a client from storage.

        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *id* (``int``) --
                Optional, whether to automatically create guest clients, defaults False
            * *token* (``aiosu.classes.token.OAuthToken``) --
                Optional, token of client to add, defaults to None

        :raises ValueError: If neither id nor token are specified
        :return: The requested client
        :rtype: aiosu.v2.client.Client
        """
        client_uid: int = kwargs.pop("id", None)
        token: OAuthToken = kwargs.pop("token", None)
        if self.client_exists(client_uid):
            return self.clients[client_uid]
        if token is not None:
            return await self.add_client(token)
        raise ValueError("Either id or token must be specified.")

    async def close(self) -> None:
        r"""Closes all client sessions."""
        for client in self.clients.values():
            await client.close()



"""
This module contains miscellaneous helper functions.
"""
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any
    from typing import Callable
    from typing import TypeVar

    T = TypeVar("T")


def from_list(f: Callable[[Any], T], x: Any) -> list[T]:
    """Applies a function to all elements in a list.

    :param f: Function to apply on list elements
    :type f: Callable[[Any], T]
    :param x: List of objects
    :type x: Any
    :raises TypeError: If x is not a list
    :return: New list
    :rtype: list[T]
    """
    if not isinstance(x, list):
        raise TypeError("Wrong type received. Expected list.")
    return [f(y) for y in x]



from __future__ import annotations

from datetime import date
from importlib import metadata

__title__ = "aiosu"
__author__ = "Nice Aesthetics"
__license__ = "GPLv3+"
__copyright__ = f"Copyright {date.today().year} Nice Aesthetics"

try:
    __version__ = metadata.version(__package__)
except metadata.PackageNotFoundError:
    import toml

    __version__ = toml.load("pyproject.toml")["tool"]["poetry"]["version"] + "dev"

from . import classes, v1, v2, utils



"""
This module handles API requests for API v2 (OAuth).

You can read more about it here: https://osu.ppy.sh/docs/index.html
"""
from __future__ import annotations

import datetime
import functools
from typing import TYPE_CHECKING

import orjson

if TYPE_CHECKING:
    from types import TracebackType
    from typing import Any
    from typing import Callable
    from typing import Optional
    from typing import Type
    from typing import Union

    from ..classes.events import BaseEvent

from ..classes import APIException
from ..classes import Beatmap
from ..classes import Score
from ..classes import Mods
from ..classes import Gamemode
from ..classes import UserQueryType
from ..classes import BeatmapDifficultyAttributes
from ..classes.events import ClientUpdateEvent
from .. import helpers
from ..classes import User
import aiohttp
from aiolimiter import AsyncLimiter
from ..classes import OAuthToken
from ..classes.events import Eventable


def check_token(func: Callable) -> Callable:
    """
    A decorator that checks the current token, to be used as:
    @check_token
    """

    @functools.wraps(func)
    async def _check_token(self: Client, *args: Any, **kwargs: Any) -> Any:
        if datetime.datetime.now() > self.token.expires_on:
            await self._refresh()
        return await func(self, *args, **kwargs)

    return _check_token


def rate_limited(func: Callable) -> Callable:
    """
    A decorator that enforces rate limiting, to be used as:
    @rate_limited
    """

    @functools.wraps(func)
    async def _rate_limited(self: Client, *args: Any, **kwargs: Any) -> Any:
        if self._session is None:
            self._session = aiohttp.ClientSession(
                headers=self._get_headers(),
            )

        async with self._limiter:
            return await func(self, *args, **kwargs)

    return _rate_limited


class Client(Eventable):
    r"""osu! API v2 Client

    :param \**kwargs:
        See below

    :Keyword Arguments:
        * *client_id* (``int``)
        * *client_secret* (``str``)
        * *base_url* (``str``) --
            Optional, base API URL, defaults to \"https://osu.ppy.sh\"
        * *token* (``aiosu.classes.token.OAuthToken``)
        * *limiter* (``aiolimiter.AsyncLimiter``) --
            Optional, custom AsyncLimiter, defaults to AsyncLimiter(1200, 60)
    """

    def __init__(self, **kwargs: Any) -> None:
        super().__init__()
        self.client_id: int = kwargs.pop("client_id", None)
        self.client_secret: str = kwargs.pop("client_secret", None)
        self.token: OAuthToken = kwargs.pop("token", OAuthToken())
        self.base_url: str = kwargs.pop("base_url", "https://osu.ppy.sh")
        self._limiter: AsyncLimiter = kwargs.pop("limiter", AsyncLimiter(1200, 60))
        self._session: aiohttp.ClientSession = None  # type: ignore

    def _get_headers(self) -> dict[str, str]:
        return {
            "Authorization": f"Bearer {self.token.access_token}",
            "Content-Type": "application/json",
            "Accept": "application/json",
        }

    async def __aenter__(self) -> Client:
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        await self.close()

    def on_client_update(self, func: Callable) -> Callable:
        """
        A decorator that is called whenever a client is updated, to be used as:

            @client.on_client_update

            async def func(event: ClientUpdateEvent)
        """
        self._listeners.append(func)

        @functools.wraps(func)
        async def _on_client_update(*args: Any, **kwargs: Any) -> Any:
            return await func(*args, **kwargs)

        return _on_client_update

    async def _process_event(self, event: BaseEvent) -> None:
        if isinstance(event, ClientUpdateEvent):
            for func in event.client._listeners:
                await func(event)
            return
        raise NotImplementedError(f"{event!r}")

    @rate_limited
    async def _refresh(self) -> None:
        """INTERNAL: Refreshes the client's token

        :raises APIException: Contains status code and error message
        """
        old_token = self.token
        url = f"{self.base_url}/oauth/token"
        headers = {"Content-Type": "application/x-www-form-urlencoded"}
        data = {
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "grant_type": "refresh_token",
            "refresh_token": old_token.refresh_token,
        }

        async with aiohttp.ClientSession(headers=headers) as temp_session:
            async with temp_session.post(url, data=data) as resp:
                try:
                    body = await resp.read()
                    json = orjson.loads(body)
                    if resp.status != 200:
                        raise APIException(resp.status, json.get("error", ""))
                    self.token = OAuthToken.parse_obj(json)
                    await self._session.close()
                    self._session = aiohttp.ClientSession(headers=self._get_headers())
                except aiohttp.client_exceptions.ContentTypeError:
                    raise APIException(403, "Invalid token specified.")

        await self._process_event(
            ClientUpdateEvent(client=self, old_token=old_token, new_token=self.token),
        )

    @rate_limited
    @check_token
    async def get_me(self, **kwargs: Any) -> User:
        r"""Gets the user who owns the current token

        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *mode* (``aiosu.classes.gamemode.Gamemode``) --
                Optional, gamemode to search for

        :raises APIException: Contains status code and error message
        :return: Requested user
        :rtype: aiosu.classes.user.User
        """
        url = f"{self.base_url}/api/v2/me"
        if "mode" in kwargs:
            mode = Gamemode(kwargs.pop("mode"))  # type: ignore
            url += f"/{mode}"
        async with self._session.get(url) as resp:
            body = await resp.read()
            json = orjson.loads(body)
            if resp.status != 200:
                raise APIException(resp.status, json.get("error", ""))
            return User.parse_obj(json)

    @rate_limited
    @check_token
    async def get_user(self, user_query: Union[str, int], **kwargs: Any) -> User:
        r"""Gets a user by a query.

        :param user_query: Username or ID to search by
        :type user_query: Union[str, int]
        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *mode* (``aiosu.classes.gamemode.Gamemode``) --
                Optional, gamemode to search for
            * *qtype* (``str``) --
                Optional, \"string\" or \"id\". Type of the user_query

        :raises APIException: Contains status code and error message
        :return: Requested user
        :rtype: aiosu.classes.user.User
        """
        url = f"{self.base_url}/api/v2/users/{user_query}"
        params = {}
        if "mode" in kwargs:
            mode = Gamemode(kwargs.pop("mode"))  # type: ignore
            url += f"/{mode}"
        if "qtype" in kwargs:
            qtype = UserQueryType(kwargs.pop("qtype"))  # type: ignore
            params["type"] = qtype.new_api_name
        async with self._session.get(url, params=params) as resp:
            body = await resp.read()
            json = orjson.loads(body)
            if resp.status != 200:
                raise APIException(resp.status, json.get("error", ""))
            return User.parse_obj(json)

    @rate_limited
    @check_token
    async def __get_type_scores(
        self, user_id: int, request_type: str, **kwargs: Any
    ) -> list[Score]:
        r"""INTERNAL: Get a user's scores by type

        :param user_id: User ID to search by
        :type user_id: int
        :param request_type: "recent", "best" or "firsts"
        :type request_type: str
        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *mode* (``aiosu.classes.gamemode.Gamemode``) --
                Optional, gamemode to search for
            * *limit* (``int``) --
                Optional, number of scores to get. Min: 1, Max: 100, defaults to 100
            * *include_fails* (``bool``) --
                Optional, whether to include failed scores, defaults to ``False``
            * *offset* (``int``) --
                Optional, page offset to start from, defaults to 0

        :raises ValueError: If limit is not between 1 and 100
        :raises ValueError: If type is invalid
        :raises APIException: Contains status code and error message
        :return: List of requested scores
        :rtype: list[aiosu.classes.score.Score]
        """
        if not 1 <= kwargs.get("limit", 100) <= 100:
            raise ValueError("Invalid limit specified. Limit must be between 1 and 100")
        if request_type not in ("recent", "best", "firsts"):
            raise ValueError(
                f'"{request_type}" is not a valid request_type. Valid options are: "recent", "best", "firsts"',
            )
        url = f"{self.base_url}/api/v2/users/{user_id}/scores/{request_type}"
        params = {
            "include_fails": kwargs.pop("include_fails", False),
            "offset": kwargs.pop("offset", 0),
            "limit": kwargs.pop("limit", 100),
        }
        if "mode" in kwargs:
            mode = Gamemode(kwargs.pop("mode"))  # type: ignore
            params["mode"] = str(mode)
        if "limit" in kwargs:
            params["limit"] = kwargs.pop("limit")
        async with self._session.get(url, params=params) as resp:
            body = await resp.read()
            json = orjson.loads(body)
            if resp.status != 200:
                raise APIException(resp.status, json.get("error", ""))
            return helpers.from_list(Score.parse_obj, json)

    async def get_user_recents(self, user_id: int, **kwargs: Any) -> list[Score]:
        r"""Get a user's recent scores.

        :param user_id: User ID to search by
        :type user_id: int
        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *mode* (``aiosu.classes.gamemode.Gamemode``) --
                Optional, gamemode to search for
            * *limit* (``int``) --
                Optional, number of scores to get. Min: 1, Max: 100, defaults to 100
            * *include_fails* (``bool``) --
                Optional, whether to include failed scores, defaults to ``False``
            * *offset* (``int``) --
                Optional, page offset to start from, defaults to 0

        :raises APIException: Contains status code and error message
        :return: List of requested scores
        :rtype: list[aiosu.classes.score.Score]
        """
        return await self.__get_type_scores(user_id, "recent", **kwargs)

    async def get_user_bests(self, user_id: int, **kwargs: Any) -> list[Score]:
        r"""Get a user's top scores.

        :param user_id: User ID to search by
        :type user_id: int
        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *mode* (``aiosu.classes.gamemode.Gamemode``) --
                Optional, gamemode to search for
            * *limit* (``int``) --
                Optional, number of scores to get. Min: 1, Max: 100, defaults to 100
            * *include_fails* (``bool``) --
                Optional, whether to include failed scores, defaults to ``False``
            * *offset* (``int``) --
                Optional, page offset to start from, defaults to 0

        :raises APIException: Contains status code and error message
        :return: List of requested scores
        :rtype: list[aiosu.classes.score.Score]
        """
        return await self.__get_type_scores(user_id, "best", **kwargs)

    async def get_user_firsts(self, user_id: int, **kwargs: Any) -> list[Score]:
        r"""Get a user's first place scores.

        :param user_id: User ID to search by
        :type user_id: int
        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *mode* (``aiosu.classes.gamemode.Gamemode``) --
                Optional, gamemode to search for
            * *limit* (``int``) --
                Optional, number of scores to get. Min: 1, Max: 100, defaults to 100
            * *include_fails* (``bool``) --
                Optional, whether to include failed scores, defaults to ``False``
            * *offset* (``int``) --
                Optional, page offset to start from, defaults to 0

        :raises APIException: Contains status code and error message
        :return: List of requested scores
        :rtype: list[aiosu.classes.score.Score]
        """
        return await self.__get_type_scores(user_id, "firsts", **kwargs)

    @rate_limited
    @check_token
    async def get_user_beatmap_scores(
        self, user_id: int, beatmap_id: int, **kwargs: Any
    ) -> list[Score]:
        r"""Get a user's scores on a specific beatmap.

        :param user_id: User ID to search by
        :type user_id: int
        :param beatmap_id: Beatmap ID to search by
        :type beatmap_id: int
        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *mode* (``aiosu.classes.gamemode.Gamemode``) --
                Optional, gamemode to search for

        :raises APIException: Contains status code and error message
        :return: List of requested scores
        :rtype: list[aiosu.classes.score.Score]
        """
        url = f"{self.base_url}/api/v2/beatmaps/{beatmap_id}/scores/users/{user_id}/all"
        params = {}
        if "mode" in kwargs:
            mode = Gamemode(kwargs.pop("mode"))  # type: ignore
            params["mode"] = str(mode)
        async with self._session.get(url) as resp:
            body = await resp.read()
            json = orjson.loads(body)
            if resp.status != 200:
                raise APIException(resp.status, json.get("error", ""))
            return helpers.from_list(Score.parse_obj, json.get("scores", []))

    @rate_limited
    @check_token
    async def get_beatmap_scores(self, beatmap_id: int, **kwargs: Any) -> list[Score]:
        r"""Get scores submitted on a specific beatmap.

        :param beatmap_id: Beatmap ID to search by
        :type beatmap_id: int
        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *mode* (``aiosu.classes.gamemode.Gamemode``) --
                Optional, gamemode to search for
            * *mods* (``aiosu.classes.mods.Mods``) --
                Optional, mods to search for
            * *type* (``str``) --
                Optional, beatmap score ranking type

        :raises APIException: Contains status code and error message
        :return: List of requested scores
        :rtype: list[aiosu.classes.score.Score]
        """
        url = f"{self.base_url}/api/v2/beatmaps/{beatmap_id}/scores"
        params = {}
        if "mode" in kwargs:
            mode = Gamemode(kwargs.pop("mode"))  # type: ignore
            params["mode"] = str(mode)
        if "mods" in kwargs:
            mods = Mods(kwargs.pop("mods"))
            params["mode"] = str(mods)
        if "type" in kwargs:
            params["type"] = kwargs.pop("type")
        async with self._session.get(url) as resp:
            body = await resp.read()
            json = orjson.loads(body)
            if resp.status != 200:
                raise APIException(resp.status, json.get("error", ""))
            return helpers.from_list(Score.parse_obj, json.get("scores", []))

    @rate_limited
    @check_token
    async def get_beatmap(self, beatmap_id: int) -> Beatmap:
        r"""Get beatmap data.

        :param beatmap_id: The ID of the beatmap
        :type beatmap_id: int
        :raises APIException: Contains status code and error message
        :return: Beatmap data object
        :rtype: aiosu.classes.beatmap.Beatmap
        """
        url = f"{self.base_url}/api/v2/beatmaps/{beatmap_id}"
        async with self._session.get(url) as resp:
            body = await resp.read()
            json = orjson.loads(body)
            if resp.status != 200:
                raise APIException(resp.status, json.get("error", ""))
            return Beatmap.parse_obj(json)

    @rate_limited
    @check_token
    async def get_beatmap_attributes(
        self, beatmap_id: int, **kwargs: Any
    ) -> BeatmapDifficultyAttributes:
        r"""Gets difficulty attributes for a beatmap.

        :param beatmap_id: The ID of the beatmap
        :type beatmap_id: int
        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *mode* (``aiosu.classes.gamemode.Gamemode``) --
                Optional, gamemode to search for
            * *mods* (``aiosu.classes.mods.Mods``) --
                Optional, mods to apply to the result

        :raises APIException: Contains status code and error message
        :return: Difficulty attributes for a beatmap
        :rtype: aiosu.classes.beatmap.BeatmapDifficultyAttributes
        """
        url = f"{self.base_url}/api/v2/beatmaps/{beatmap_id}/attributes"
        params: dict[str, Any] = {}
        if "mode" in kwargs:
            mode = Gamemode(kwargs.pop("mode"))  # type: ignore
            params["ruleset_id"] = int(mode)
        if "mods" in kwargs:
            mods = Mods(kwargs.pop("mods"))
            params["mods"] = str(mods)
        async with self._session.post(url, data=params) as resp:
            body = await resp.read()
            json = orjson.loads(body)
            if resp.status != 200:
                raise APIException(resp.status, json.get("error", ""))
            return BeatmapDifficultyAttributes.parse_obj(json.get("attributes"))

    async def close(self) -> None:
        """Closes the client session."""
        if self._session:
            await self._session.close()



"""
This module contains accuracy calculators for osu! gamemodes.
"""
from __future__ import annotations

import abc
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..classes import Score

__all__ = [
    "OsuAccuracyCalculator",
    "TaikoAccuracyCalculator",
    "ManiaAccuracyCalculator",
    "CatchAccuracyCalculator",
]


class AbstractAccuracyCalculator(abc.ABC):
    @staticmethod
    @abc.abstractmethod
    def calculate(score: Score) -> float:
        ...

    @staticmethod
    @abc.abstractmethod
    def calculate_weighted(score: Score) -> float:
        ...


class OsuAccuracyCalculator(AbstractAccuracyCalculator):
    @staticmethod
    def calculate(score: Score) -> float:
        """Calculates accuracy for an osu!std score.

        :param score: The score to calculate accuracy for
        :type score: aiosu.classes.score.Score
        :return: Accuracy for the given score
        :rtype: float
        """
        total_hits = (
            score.statistics.count_300
            + score.statistics.count_100
            + score.statistics.count_50
            + score.statistics.count_miss
        )

        accuracy = 0.0
        if total_hits > 0:
            accuracy = max(
                (
                    score.statistics.count_300 * 6
                    + score.statistics.count_100 * 2
                    + score.statistics.count_50
                )
                / (total_hits * 6),
                0,
            )

        return accuracy

    @staticmethod
    def calculate_weighted(score: Score) -> float:
        """Calculates weighted accuracy for an osu!std score.

        :param score: The score to calculate accuracy for
        :type score: aiosu.classes.score.Score
        :raises ValueError: If score does not have an associated beatmap
        :raises ValueError: If the associated beatmap does not have object counts
        :return: Weighted accuracy to be used in pp calculation
        :rtype: float
        """
        if score.beatmap is None:
            raise ValueError("Given score does not have a beatmap.")

        total_hits = (
            score.statistics.count_300
            + score.statistics.count_100
            + score.statistics.count_50
            + score.statistics.count_miss
        )

        if (amount_hit_objects_with_accuracy := score.beatmap.count_circles) is None:
            raise ValueError("Beatmap object does not contain object information.")

        better_accuracy_percentage = 0.0
        if amount_hit_objects_with_accuracy > 0:
            better_accuracy_percentage = max(
                (
                    (
                        score.statistics.count_300
                        - (total_hits - amount_hit_objects_with_accuracy)
                    )
                    * 6
                    + score.statistics.count_100 * 2
                    + score.statistics.count_50
                )
                / (amount_hit_objects_with_accuracy * 6),
                0,
            )

        return better_accuracy_percentage


class TaikoAccuracyCalculator(AbstractAccuracyCalculator):
    @staticmethod
    def calculate(score: Score) -> float:
        """Calculates accuracy for an osu!taiko score.

        :param score: The score to calculate accuracy for
        :type score: aiosu.classes.score.Score
        :return: Accuracy for the given score
        :rtype: float
        """
        total_hits = (
            score.statistics.count_300
            + score.statistics.count_100
            + score.statistics.count_50
            + score.statistics.count_miss
        )

        accuracy = 0.0
        if total_hits > 0:
            accuracy = (
                score.statistics.count_300 * 2.0 + score.statistics.count_100
            ) / (total_hits * 2.0)

        return accuracy

    @classmethod
    def calculate_weighted(cls, score: Score) -> float:
        """Calculates weighted accuracy for an osu!taiko score.

        :param score: The score to calculate accuracy for
        :type score: aiosu.classes.score.Score
        :return: Weighted accuracy to be used in pp calculation
        :rtype: float
        """
        return cls.calculate(score)


class ManiaAccuracyCalculator(AbstractAccuracyCalculator):
    @staticmethod
    def calculate(score: Score) -> float:
        """Calculates accuracy for an osu!mania score.

        :param score: The score to calculate accuracy for
        :type score: aiosu.classes.score.Score
        :return: Accuracy for the given score
        :rtype: float
        """
        count_perfect = score.statistics.count_geki
        count_great = score.statistics.count_300
        count_good = score.statistics.count_katu
        count_ok = score.statistics.count_100
        count_meh = score.statistics.count_50
        count_miss = score.statistics.count_miss

        total_hits = (
            count_perfect + count_ok + count_great + count_good + count_meh + count_miss
        )

        accuracy = 0.0
        if total_hits > 0:
            accuracy = (
                +((count_perfect + count_great) * 300)
                + (count_good * 200)
                + (count_ok * 100)
                + (count_meh * 50)
            ) / (total_hits * 300)

        return accuracy

    @staticmethod
    def calculate_weighted(score: Score) -> float:
        """Calculates weighted accuracy for an osu!mania score.

        :param score: The score to calculate accuracy for
        :type score: aiosu.classes.score.Score
        :return: Weighted accuracy to be used in pp calculation
        :rtype: float
        """
        count_perfect = score.statistics.count_geki
        count_great = score.statistics.count_300
        count_good = score.statistics.count_katu
        count_ok = score.statistics.count_100
        count_meh = score.statistics.count_50
        count_miss = score.statistics.count_miss

        total_hits = (
            count_perfect + count_ok + count_great + count_good + count_meh + count_miss
        )

        accuracy = 0.0
        if total_hits > 0:
            accuracy = (
                +(count_perfect * 320)
                + (count_great * 300)
                + (count_good * 200)
                + (count_ok * 100)
                + (count_meh * 50)
            ) / (total_hits * 320)

        return accuracy


class CatchAccuracyCalculator(AbstractAccuracyCalculator):
    @staticmethod
    def calculate(score: Score) -> float:
        """Calculates accuracy for an osu!catch score.

        :param score: The score to calculate accuracy for
        :type score: aiosu.classes.score.Score
        :return: Accuracy for the given score
        :rtype: float
        """
        fruits_hit = score.statistics.count_300
        ticks_hit = score.statistics.count_100
        tiny_ticks_hit = score.statistics.count_50
        tiny_ticks_missed = score.statistics.count_katu
        misses = score.statistics.count_miss

        total_combo_hits = misses + ticks_hit + fruits_hit
        total_hits = (
            tiny_ticks_hit + ticks_hit + fruits_hit + misses + tiny_ticks_missed
        )
        successful_hits = tiny_ticks_hit + ticks_hit + fruits_hit

        accuracy = 0.0
        if total_hits != 0:
            accuracy = float(successful_hits) / total_hits

        return accuracy

    @classmethod
    def calculate_weighted(cls, score: Score) -> float:
        """Calculates weighted accuracy for an osu!catch score.

        :param score: The score to calculate accuracy for
        :type score: aiosu.classes.score.Score
        :return: Weighted accuracy to be used in pp calculation
        :rtype: float
        """
        return cls.calculate(score)



"""
This module contains models for Beatmap objects.
"""
from __future__ import annotations

import datetime
from enum import Enum
from typing import Optional
from typing import TYPE_CHECKING

from pydantic import Field
from pydantic import root_validator

from .gamemode import Gamemode
from .models import BaseModel

if TYPE_CHECKING:
    from typing import Any


class BeatmapRankStatus(Enum):
    GRAVEYARD = (-2, "graveyard")
    WIP = (-1, "wip")
    PENDING = (0, "pending")
    RANKED = (1, "ranked")
    APPROVED = (2, "approved")
    QUALIFIED = (3, "qualified")
    LOVED = (4, "loved")

    def __init__(self, id: int, name_api: str) -> None:
        self.id: int = id
        self.name_api: str = name_api

    def __int__(self) -> int:
        return self.id

    def __str__(self) -> str:
        return self.name_api

    @classmethod
    def _missing_(cls, query: object) -> BeatmapRankStatus:
        if isinstance(query, int):
            for status in list(BeatmapRankStatus):
                if status.id == query:
                    return status
        elif isinstance(query, str):
            for status in list(BeatmapRankStatus):
                if status.name_api == query.lower():
                    return status
        raise ValueError(f"BeatmapRankStatus {query} does not exist.")


class BeatmapAvailability(BaseModel):
    more_information: Optional[str] = None
    download_disabled: Optional[bool] = None

    @classmethod
    def _from_api_v1(cls, data: Any) -> BeatmapAvailability:
        return cls.parse_obj({"download_disabled": data["download_unavailable"]})


class BeatmapNominations(BaseModel):
    current: Optional[int] = None
    required: Optional[int] = None


class BeatmapCovers(BaseModel):
    cover: str
    card: str
    list: str
    slimcover: str
    cover_2_x: Optional[str]
    card_2_x: Optional[str]
    list_2_x: Optional[str]
    slimcover_2_x: Optional[str]

    @classmethod
    def _from_api_v1(cls, data: Any) -> BeatmapCovers:
        base_url = "https://assets.ppy.sh/beatmaps/"
        return cls.parse_obj(
            {
                "cover": f"{base_url}{data['beatmapset_id']}/covers/cover.jpg",
                "card": f"{base_url}{data['beatmapset_id']}/covers/card.jpg",
                "list": f"{base_url}{data['beatmapset_id']}/covers/list.jpg",
                "slimcover": f"{base_url}{data['beatmapset_id']}/covers/slimcover.jpg",
                "cover_2_x": f"{base_url}{data['beatmapset_id']}/covers/cover@2x.jpg",
                "card_2_x": f"{base_url}{data['beatmapset_id']}/covers/card@2x.jpg",
                "list_2_x": f"{base_url}{data['beatmapset_id']}/covers/list@2x.jpg",
                "slimcover_2_x": f"{base_url}{data['beatmapset_id']}/covers/slimcover@2x.jpg",
            },
        )


class BeatmapHype(BaseModel):
    current: int
    required: int


class BeatmapFailtimes(BaseModel):
    exit: Optional[list[int]] = None
    fail: Optional[list[int]] = None


class BeatmapDifficultyAttributes(BaseModel):
    max_combo: int
    star_rating: float
    # osu standard
    aim_difficulty: Optional[float] = None
    approach_rate: Optional[float] = None  # osu catch + standard
    flashlight_difficulty: Optional[float] = None
    overall_difficulty: Optional[float] = None
    slider_factor: Optional[float] = None
    speed_difficulty: Optional[float] = None
    speed_note_count: Optional[float] = None
    # osu taiko
    stamina_difficulty: Optional[float] = None
    rhythm_difficulty: Optional[float] = None
    colour_difficulty: Optional[float] = None
    # osu mania
    great_hit_window: Optional[float] = None
    score_multiplier: Optional[float] = None


class Beatmap(BaseModel):
    id: int
    url: str
    mode: Gamemode
    beatmapset_id: int
    difficulty_rating: float
    status: BeatmapRankStatus
    total_length: int
    user_id: int
    version: str
    accuracy: Optional[float] = None
    ar: Optional[float] = None
    cs: Optional[float] = None
    bpm: Optional[float] = None
    convert: Optional[bool] = None
    count_circles: Optional[int] = None
    count_sliders: Optional[int] = None
    count_spinners: Optional[int] = None
    deleted_at: Optional[datetime.datetime] = None
    drain: Optional[float] = None
    hit_length: Optional[int] = None
    is_scoreable: Optional[bool] = None
    last_updated: Optional[datetime.datetime] = None
    passcount: Optional[int] = None
    play_count: Optional[int] = Field(None, alias="playcount")
    checksum: Optional[str] = None
    max_combo: Optional[int] = None
    beatmapset: Optional[Beatmapset] = None
    failtimes: Optional[BeatmapFailtimes] = None

    @root_validator(pre=True)
    def _set_url(cls, values: dict[str, Any]) -> dict[str, Any]:
        if values.get("url") is None:
            id = values["id"]
            beatmapset_id = values["beatmapset_id"]
            mode = Gamemode(values["mode"])  # type: ignore
            values[
                "url"
            ] = f"https://osu.ppy.sh/beatmapsets/{beatmapset_id}#{mode}/{id}"
        return values

    @property
    def discussion_url(self) -> str:
        return f"https://osu.ppy.sh/beatmapsets/{self.beatmapset_id}/discussion/{self.id}/general"

    @property
    def count_objects(self) -> int:
        """Total count of the objects.

        :raises ValueError: Raised if object counts are none
        :return: Sum of counts of all objects
        :rtype: int
        """
        if (
            self.count_circles is None
            or self.count_spinners is None
            or self.count_sliders is None
        ):
            raise ValueError("Beatmap contains no object count information.")
        return self.count_spinners + self.count_circles + self.count_sliders

    @classmethod
    def _from_api_v1(cls, data: Any) -> Beatmap:
        return cls.parse_obj(
            {
                "beatmapset_id": data["beatmapset_id"],
                "difficulty_rating": data["difficultyrating"],
                "id": data["beatmap_id"],
                "mode": int(data["mode"]),
                "status": int(data["approved"]),
                "total_length": data["total_length"],
                "hit_length": data["total_length"],
                "user_id": data["creator_id"],
                "version": data["version"],
                "accuracy": data["diff_overall"],
                "cs": data["diff_size"],
                "ar": data["diff_approach"],
                "drain": data["diff_drain"],
                "last_updated": data["last_update"],
                "bpm": data["bpm"],
                "checksum": data["file_md5"],
                "playcount": data["playcount"],
                "passcount": data["passcount"],
                "count_circles": data["count_normal"],
                "count_sliders": data["count_slider"],
                "count_spinners": data["count_spinner"],
                "max_combo": data["max_combo"],
            },
        )


class Beatmapset(BaseModel):
    id: int
    artist: str
    artist_unicode: str
    covers: BeatmapCovers
    creator: str
    favourite_count: int
    play_count: int = Field(alias="playcount")
    preview_url: str
    source: str
    status: BeatmapRankStatus
    title: str
    title_unicode: str
    user_id: int
    video: bool
    nsfw: Optional[bool] = None
    hype: Optional[BeatmapHype] = None
    availability: Optional[BeatmapAvailability] = None
    bpm: Optional[float] = None
    can_be_hyped: Optional[bool] = None
    discussion_enabled: Optional[bool] = None
    discussion_locked: Optional[bool] = None
    is_scoreable: Optional[bool] = None
    last_updated: Optional[datetime.datetime] = None
    legacy_thread_url: Optional[str] = None
    nominations_summary: Optional[BeatmapNominations] = None
    ranked_date: Optional[datetime.datetime] = None
    storyboard: Optional[bool] = None
    submitted_date: Optional[datetime.datetime] = None
    tags: Optional[str] = None
    ratings: Optional[list[int]] = None
    has_favourited: Optional[bool] = None
    beatmaps: Optional[list[Beatmap]] = None

    @property
    def url(self) -> str:
        return f"https://osu.ppy.sh/beatmapsets/{self.id}"

    @property
    def discussion_url(self) -> str:
        return f"https://osu.ppy.sh/beatmapsets/{self.id}/discussion"

    @classmethod
    def _from_api_v1(cls, data: Any) -> Beatmapset:
        return cls.parse_obj(
            {
                "id": data["beatmapset_id"],
                "artist": data["artist"],
                "artist_unicode": data["artist"],
                "covers": BeatmapCovers._from_api_v1(data),
                "favourite_count": data["favourite_count"],
                "creator": data["creator"],
                "play_count": data["playcount"],
                "preview_url": f"https://b.ppy.sh/preview/{data['beatmapset_id']}.mp3",
                "source": data["source"],
                "status": int(data["approved"]),
                "title": data["title"],
                "title_unicode": data["title"],
                "user_id": data["creator_id"],
                "video": data["video"],
                "submitted_date": data["submit_date"],
                "ranked_date": data["approved_date"],
                "last_updated": data["last_update"],
                "tags": data["tags"],
                "storyboard": data["storyboard"],
                "availabiliy": BeatmapAvailability._from_api_v1(data),
                "beatmaps": [Beatmap._from_api_v1(data)],
            },
        )


Beatmap.update_forward_refs()



"""
This module contains models for User objects.
"""
from __future__ import annotations

import datetime
from enum import Enum
from typing import Any
from typing import Optional
from typing import TYPE_CHECKING

from .common import Achievement
from .common import Country
from .common import TimestampedCount
from .gamemode import Gamemode
from .models import BaseModel

if TYPE_CHECKING:
    from typing import Callable

cast_int: Callable[..., int] = lambda x: int(x or 0)
cast_float: Callable[..., float] = lambda x: float(x or 0)


class UserQueryType(Enum):
    ID = ("id", "id")
    USERNAME = ("string", "username")

    def __init__(self, old: str, new: str) -> None:
        self.old_api_name = old
        self.new_api_name = new

    @classmethod
    def _missing_(cls, query: object) -> UserQueryType:
        for q in list(UserQueryType):
            if query in q.value:
                return q
        raise ValueError(f"UserQueryType {query} does not exist.")


class Userpage(BaseModel):
    html: str
    raw: Optional[str] = None


class UserLevel(BaseModel):
    current: int
    progress: float

    @classmethod
    def _from_api_v1(cls, data: Any) -> UserLevel:
        level = cast_float(data["level"])
        current = int(level)
        progress = (level - current) * 100
        return cls.parse_obj({"current": current, "progress": progress})


class UserKudosu(BaseModel):
    total: int
    available: int


class UserRankHistoryElement(BaseModel):
    mode: str
    data: list[int]

    @property
    def average_gain(self) -> float:
        """Average rank gain.

        :return: Average rank gain for a user
        :rtype: float
        """
        return (self.data[1] - self.data[-1]) / len(self.data)


class UserProfileCover(BaseModel):
    url: str
    custom_url: Optional[str] = None
    id: Optional[str] = None


class UserProfileTournamentBanner(BaseModel):
    tournament_id: int
    image: str
    id: Optional[int] = None


class UserBadge(BaseModel):
    awarded_at: datetime.datetime
    description: str
    image_url: str
    url: str


class UserAccountHistory(BaseModel):
    id: int
    description: str
    length: int
    timestamp: datetime.datetime
    type: str


class UserGradeCounts(BaseModel):
    ssh: int
    """Number of Silver SS ranks achieved."""
    ss: int
    """Number of SS ranks achieved."""
    sh: int
    """Number of Silver S ranks achieved."""
    s: int
    """Number of S ranks achieved."""
    a: int
    """Number of A ranks achieved."""

    @classmethod
    def _from_api_v1(cls, data: Any) -> UserGradeCounts:
        return cls.parse_obj(
            {
                "ss": cast_int(data["count_rank_ss"]),
                "ssh": cast_int(data["count_rank_ssh"]),
                "s": cast_int(data["count_rank_s"]),
                "sh": cast_int(data["count_rank_sh"]),
                "a": cast_int(data["count_rank_a"]),
            },
        )


class UserGroup(BaseModel):
    id: int
    identifier: str
    name: str
    short_name: str
    description: str
    colour: str
    is_probationary: Optional[bool] = None


class UserStats(BaseModel):
    level: UserLevel
    pp: float
    ranked_score: int
    hit_accuracy: float
    play_count: int
    play_time: int
    total_score: int
    total_hits: int
    is_ranked: bool
    grade_counts: UserGradeCounts
    replays_watched_by_others: Optional[int]
    maximum_combo: Optional[int]
    global_rank: Optional[int]
    country_rank: Optional[int]
    user: Optional[User] = None

    @classmethod
    def _from_api_v1(cls, data: Any) -> UserStats:
        """Some fields can be None, we want to force them to cast to a value."""
        return cls.parse_obj(
            {
                "level": UserLevel._from_api_v1(data),
                "pp": cast_float(data["pp_raw"]),
                "global_rank": cast_int(data["pp_rank"]),
                "country_rank": cast_int(data["pp_country_rank"]),
                "ranked_score": cast_int(data["ranked_score"]),
                "hit_accuracy": cast_float(data["accuracy"]),
                "play_count": cast_int(data["playcount"]),
                "play_time": cast_int(data["total_seconds_played"]),
                "total_score": cast_int(data["total_score"]),
                "total_hits": cast_int(data["count300"])
                + cast_int(data["count100"])
                + cast_int(data["count50"]),
                "is_ranked": cast_float(data["pp_raw"]) != 0,
                "grade_counts": UserGradeCounts._from_api_v1(data),
            },
        )


class User(BaseModel):
    avatar_url: str
    country_code: str
    id: int
    username: str
    default_group: Optional[str]
    is_active: Optional[bool]
    is_bot: Optional[bool]
    is_online: Optional[bool]
    is_supporter: Optional[bool]
    pm_friends_only: Optional[bool]
    profile_colour: Optional[str]
    is_deleted: Optional[bool] = None
    last_visit: Optional[datetime.datetime] = None
    discord: Optional[str] = None
    has_supported: Optional[bool] = None
    interests: Optional[str] = None
    join_date: Optional[datetime.datetime] = None
    kudosu: Optional[UserKudosu] = None
    location: Optional[str] = None
    max_blocks: Optional[int] = None
    max_friends: Optional[int] = None
    occupation: Optional[str] = None
    playmode: Optional[Gamemode] = None
    playstyle: Optional[list[str]] = None
    post_count: Optional[int] = None
    profile_order: Optional[list[str]] = None
    title: Optional[str] = None
    twitter: Optional[str] = None
    website: Optional[str] = None
    country: Optional[Country] = None
    cover: Optional[UserProfileCover] = None
    is_restricted: Optional[bool] = None
    account_history: Optional[list[Any]] = None  # Unsure what this is
    active_tournament_banner: Optional[UserProfileTournamentBanner] = None
    badges: Optional[list[UserBadge]] = None
    beatmap_playcounts_count: Optional[int] = None
    favourite_beatmapset_count: Optional[int] = None
    follower_count: Optional[int] = None
    graveyard_beatmapset_count: Optional[int] = None
    groups: Optional[list[UserGroup]] = None
    loved_beatmapset_count: Optional[int] = None
    monthly_playcounts: Optional[list[TimestampedCount]] = None
    page: Optional[Userpage] = None
    pending_beatmapset_count: Optional[int] = None
    previous_usernames: Optional[list[str]] = None
    ranked_beatmapset_count: Optional[int] = None
    replays_watched_counts: Optional[list[TimestampedCount]] = None
    scores_best_count: Optional[int] = None
    scores_first_count: Optional[int] = None
    scores_recent_count: Optional[int] = None
    statistics: Optional[UserStats] = None
    support_level: Optional[int] = None
    user_achievements: Optional[list[Achievement]] = None
    rank_history: Optional[UserRankHistoryElement] = None

    @property
    def url(self) -> str:
        return f"https://osu.ppy.sh/users/{self.id}"

    @classmethod
    def _from_api_v1(cls, data: Any) -> User:
        return cls.parse_obj(
            {
                "avatar_url": f"https://s.ppy.sh/a/{data['user_id']}",
                "country_code": data["country"],
                "id": data["user_id"],
                "username": data["username"],
                "join_date": data["join_date"],
                "statistics": UserStats._from_api_v1(data),
            },
        )



"""
This module contains models for mods.
"""
from __future__ import annotations

from collections import UserList
from enum import Enum
from functools import reduce
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Generator

    from typing import Any
    from typing import Union


class Mod(Enum):
    """Bitwise Flags representing osu! mods."""

    NoMod = (0, "")
    NoFail = (1, "NF")
    Easy = (2, "EZ")
    TouchDevice = (4, "TD")
    Hidden = (8, "HD")
    HardRock = (16, "HR")
    SuddenDeath = (32, "SD")
    DoubleTime = (64, "DT")
    Relax = (128, "RX")
    HalfTime = (256, "HT")
    Nightcore = (512, "NC")  # Only set along with DoubleTime. i.e: NC only gives 576
    Flashlight = (1024, "FL")
    Autoplay = (2048, "AT")
    SpunOut = (4096, "SO")
    Autopilot = (8192, "AP")  # Called Relax2 on osu! API documentation
    Perfect = (16384, "PF")  # Only set along with SuddenDeath. i.e: PF only gives 16416
    Key4 = (32768, "4K")
    Key5 = (65536, "5K")
    Key6 = (131072, "6K")
    Key7 = (262144, "7K")
    Key8 = (524288, "8K")
    FadeIn = (1048576, "FI")
    Random = (2097152, "RD")
    Key9 = (16777216, "9K")
    KeyCoop = (33554432, "CK")
    Key1 = (67108864, "1K")
    Key3 = (134217728, "3K")
    Key2 = (268435456, "2K")
    Mirror = (1073741824, "MR")

    def __init__(self, value: int, short_name: str = "") -> None:
        self.bitmask = value
        self.short_name = short_name

    def __int__(self) -> int:
        return self.bitmask

    def __str__(self) -> str:
        return self.short_name

    def __and__(self, __o: int) -> int:
        return int(self) & __o

    def __or__(self, __o: int) -> int:
        return int(self) | __o

    @classmethod
    def from_type(cls, __o: object) -> Mod:
        if isinstance(__o, cls):
            return __o
        for mod in list(Mod):
            if __o in mod.value:
                return mod
        raise ValueError(f"Mod {__o!r} does not exist.")

    @classmethod
    def _missing_(cls, query: object) -> Mod:
        return cls.from_type(query)


class Mods(UserList):
    """List of Mod objects"""

    def __init__(self, mods: Union[list[str], str, int] = []) -> None:
        super().__init__(self)
        self.data = []
        if isinstance(mods, str):  # string of mods
            mods = [mods[i : i + 2] for i in range(0, len(mods), 2)]
        if isinstance(mods, list):  # List of Mod types
            self.data = [Mod(mod) for mod in mods]  # type: ignore
        if isinstance(mods, int):  # Bitwise representation of mods
            self.data = [mod for mod in list(Mod) if mod & mods]

    @property
    def bitwise(self) -> int:
        """Bitwise representation.

        :return: Bitwise representation of the mod combination
        :rtype: int
        """
        return reduce(lambda x, y: int(x) | int(y), self, 0)

    def __str__(self) -> str:
        if len(self) == 0:
            return "NM"

        result: str = ""
        for mod in self:
            if Mod.Nightcore in self and mod is Mod.DoubleTime:
                continue
            if Mod.Perfect in self and mod is Mod.SuddenDeath:
                continue

            result += mod.short_name
        return result

    def __int__(self) -> int:
        return self.bitwise

    def __and__(self, __o: Any) -> int:
        if isinstance(__o, int):
            return int(self) & __o
        if isinstance(__o, Mod):
            return int(self) & int(__o)
        if isinstance(__o, Mods):
            return int(self) & int(__o)
        raise ValueError(f"Object {__o!r} is of invalid type.")

    def __or__(self, __o: Any) -> int:
        if isinstance(__o, int):
            return int(self) | __o
        if isinstance(__o, Mod):
            return int(self) | int(__o)
        if isinstance(__o, Mods):
            return int(self) | int(__o)
        raise ValueError(f"Object {__o!r} is of invalid type.")

    @classmethod
    def __get_validators__(cls) -> Generator:
        yield cls._validate

    @classmethod
    def __modify_schema__(cls, field_schema):  # type: ignore
        pass  # Genuinely not sure about implementing this

    @classmethod
    def _validate(cls, v: object) -> Mods:
        if not isinstance(v, (list, str, int)):
            raise TypeError("Invalid type specified ")
        return cls(v)



"""
This module contains models for miscellaneous objects.
"""
from __future__ import annotations

import datetime

from emojiflags.lookup import lookup as flag_lookup  # type: ignore
from pydantic import validator

from .models import BaseModel


class TimestampedCount(BaseModel):
    start_date: datetime.datetime
    count: int

    @validator("start_date", pre=True)
    def _date_validate(cls, v: str) -> datetime.datetime:
        return datetime.datetime.strptime(v, "%Y-%m-%d")


class Achievement(BaseModel):
    achieved_at: datetime.datetime
    achievement_id: int


class Country(BaseModel):
    code: str
    name: str

    @property
    def flag_emoji(self) -> str:
        """Emoji for the flag.

        :return: Unicode emoji representation of the country's flag
        :rtype: str
        """
        return flag_lookup(self.code)



from __future__ import annotations

import datetime
from enum import Enum
from typing import Optional
from typing import TYPE_CHECKING

from pydantic import Field
from pydantic import root_validator
from pydantic import validator

from ..gamemode import Gamemode
from ..models import BaseModel
from ..mods import Mods
from ..score import ScoreStatistics

if TYPE_CHECKING:
    from typing import Any


class MatchTeam(Enum):
    NONE = 0
    BLUE = 1
    RED = 2


class MatchScoringType(Enum):
    SCORE = 0
    ACCURACY = 1
    COMBO = 2
    SCOREV2 = 3


class MatchTeamType(Enum):
    HEADTOHEAD = 0
    COOP = 1
    TEAMVS = 2
    TAGTEAMVS = 3


class MatchScore(BaseModel):
    slot: int
    team: MatchTeam
    user_id: int
    score: int
    max_combo: int = Field(alias="maxcombo")
    passed: bool = Field(alias="pass")
    perfect: bool
    enabled_mods: Mods
    statistics: ScoreStatistics

    def get_full_mods(self, game: MatchGame) -> Mods:
        """Get the mods including globals from the match.

        :param game: The match the score took place in
        :type game: MatchGame
        :return: Mods list of Mod objects
        :rtype: Mods
        """
        return Mods(self.enabled_mods | game.mods)

    @root_validator(pre=True)
    def _set_statistics(cls, values: dict[str, Any]) -> dict[str, Any]:
        values["statistics"] = ScoreStatistics._from_api_v1(values)
        return values

    @validator("enabled_mods", pre=True)
    def _set_enabled_mods(cls, v: Any) -> int:
        if v is not None:
            return int(v)
        return 0

    @validator("team", pre=True)
    def _set_team(cls, v: Any) -> int:
        return int(v)


class MatchGame(BaseModel):
    """Multiplayer game API object."""

    id: int = Field(alias="game_id")
    start_time: datetime.datetime
    beatmap_id: int
    mode: Gamemode = Field(alias="play_mode")
    match_type: int
    scoring_type: MatchScoringType
    team_type: MatchTeamType
    scores: list[MatchScore]
    mods: Mods
    end_time: Optional[datetime.datetime]
    """None if game was aborted."""

    @validator("mode", pre=True)
    def _set_mode(cls, v: Any) -> int:
        return int(v)

    @validator("mods", pre=True)
    def _set_mods(cls, v: Any) -> int:
        if v is not None:
            return int(v)
        return 0

    @validator("scoring_type", pre=True)
    def _set_scoring_type(cls, v: Any) -> int:
        return int(v)

    @validator("team_type", pre=True)
    def _set_team_type(cls, v: Any) -> int:
        return int(v)


class Match(BaseModel):
    """Multiplayer match API object."""

    id: int = Field(alias="match_id")
    name: str
    start_time: datetime.datetime
    games: list[MatchGame]
    end_time: Optional[datetime.datetime]
    """None if game is ongoing."""

    @root_validator(pre=True)
    def _format_values(cls, values: dict[str, Any]) -> dict[str, Any]:
        return {**values["match"], "games": values["games"]}



"""
This module contains models for API v2 token objects.
"""
from __future__ import annotations

import datetime
from typing import TYPE_CHECKING

from pydantic import root_validator

from .models import BaseModel

if TYPE_CHECKING:
    from typing import Any


class OAuthToken(BaseModel):
    token_type: str = "Bearer"
    """Defaults to 'Bearer'"""
    access_token: str = ""
    refresh_token: str = ""
    expires_on: datetime.datetime = datetime.datetime.utcfromtimestamp(0)
    """Can be a datetime.datetime object or a string. Alternatively, expires_in may be passed representing the number of seconds the token will be valid for."""

    @root_validator(pre=True)
    def _set_expires_on(cls, values: dict[str, Any]) -> dict[str, Any]:
        if isinstance(values.get("expires_in"), int):
            values["expires_on"] = datetime.datetime.utcnow() + datetime.timedelta(
                seconds=values["expires_in"],
            )
        return values



from __future__ import annotations

from .match import Match
from .match import MatchGame
from .match import MatchScore
from .match import MatchScoringType
from .match import MatchTeam
from .match import MatchTeamType
from .replay import Replay



"""
This module contains models for Gamemode objects.
"""
from __future__ import annotations

from enum import Enum


class Gamemode(Enum):
    STANDARD = (0, "https://i.imgur.com/lT2nqls.png", "Standard", "STD", "osu")
    TAIKO = (1, "https://i.imgur.com/G6bzM0X.png", "Taiko", "Taiko", "taiko")
    CTB = (2, "https://i.imgur.com/EsanYkH.png", "Catch the Beat", "CTB", "fruits")
    MANIA = (3, "https://i.imgur.com/0uZM1PZ.png", "Mania", "Mania", "mania")

    def __init__(
        self,
        id: int,
        icon: str,
        name_full: str,
        name_short: str,
        name_api: str,
    ) -> None:
        self.id: int = id
        self.icon: str = icon
        self.name_full: str = name_full
        self.name_short: str = name_short
        self.name_api: str = name_api

    def __int__(self) -> int:
        return self.id

    def __str__(self) -> str:
        return self.name_api

    def __format__(self, spec: str) -> str:
        if spec == "f":
            return self.name_full
        if spec == "s":
            return self.name_short
        return self.name_api

    @classmethod
    def from_type(cls, __o: object) -> Gamemode:
        """Gets a gamemode.

        :param __o: Object to search for
        :type __o: object
        :raises ValueError: If object cannot be converted to Gamemode
        :return: A Gamemode object. Will search by name_api, name_short, id
        :rtype: aiosu.classes.gamemode.Gamemode
        """
        if isinstance(__o, cls):
            return __o
        for mode in list(Gamemode):
            if __o in (mode.name_api, mode.name_short, mode.id):
                return mode
        raise ValueError(f"Gamemode {__o} does not exist.")

    @classmethod
    def _missing_(cls, query: object) -> Gamemode:
        return cls.from_type(query)



from __future__ import annotations

from . import events
from .beatmap import Beatmap
from .beatmap import BeatmapAvailability
from .beatmap import BeatmapCovers
from .beatmap import BeatmapDifficultyAttributes
from .beatmap import BeatmapFailtimes
from .beatmap import BeatmapHype
from .beatmap import BeatmapNominations
from .beatmap import BeatmapRankStatus
from .beatmap import Beatmapset
from .common import Achievement
from .common import Country
from .common import TimestampedCount
from .exceptions import APIException
from .gamemode import Gamemode
from .models import BaseModel
from .models import FrozenModel
from .mods import Mod
from .mods import Mods
from .performance import CatchPerformanceAttributes
from .performance import ManiaPerformanceAttributes
from .performance import OsuPerformanceAttributes
from .performance import PerformanceAttributes
from .performance import TaikoPerformanceAttributes
from .score import Score
from .score import ScoreStatistics
from .score import ScoreWeight
from .token import OAuthToken
from .user import User
from .user import UserAccountHistory
from .user import UserBadge
from .user import UserGradeCounts
from .user import UserGroup
from .user import UserKudosu
from .user import UserLevel
from .user import Userpage
from .user import UserProfileCover
from .user import UserProfileTournamentBanner
from .user import UserQueryType
from .user import UserRankHistoryElement
from .user import UserStats



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aiosu',
 'aiosu.classes',
 'aiosu.classes.legacy',
 'aiosu.utils',
 'aiosu.v1',
 'aiosu.v2']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.8.3,<4.0.0',
 'aiolimiter>=1.0.0,<2.0.0',
 'emojiflags>=0.1.1,<0.2.0',
 'orjson>=3.8.3,<4.0.0',
 'pydantic>=1.10.2,<2.0.0']

extras_require = \
{'docs': ['toml>=0.10.2,<0.11.0',
          'sphinx>=5.3.0,<6.0.0',
          'sphinx-rtd-theme>=1.1.1,<2.0.0'],
 'test': ['pytest>=7.2.0,<8.0.0',
          'pytest-asyncio>=0.20.2,<0.21.0',
          'pytest-mock>=3.10.0,<4.0.0',
          'toml>=0.10.2,<0.11.0',
          'types-toml>=0.10.8.1,<0.11.0.0']}

setup_kwargs = {
    'name': 'aiosu',
    'version': '0.1.3',
    'description': 'Simple and fast osu! API v1 and v2 library',
    'long_description': 'aiosu\n=====\n\n|Python| |pypi| |pre-commit.ci status| |rtd| |pytest| |mypy|\n\nSimple and fast osu! API v1 and v2 library\n\n\nFeatures\n--------\n\n- Support for API v1 and API v2\n- Rate limit handling\n- Utilities for osu! related calculations\n- Easy to use\n\n\nInstalling\n----------\n\n**Python 3.9 or higher is required**\n\nTo install the library, simply run the following commands\n\n.. code:: sh\n\n    # Linux/macOS\n    python3 -m pip install -U aiosu\n\n    # Windows\n    py -3 -m pip install -U aiosu\n\nTo install the development version, do the following:\n\n.. code:: sh\n\n    $ git clone https://github.com/NiceAesth/aiosu\n    $ cd aiosu\n    $ python3 -m pip install -U .\n\n\nAPI v1 Example\n--------------\n\n.. code:: py\n\n   import aiosu\n   import asyncio\n\n\n   async def main():\n       # async with syntax\n       async with aiosu.v1.Client("osu api token") as client:\n           user = await client.get_user(7782553)\n\n       # regular syntax\n       client = aiosu.v1.Client("osu api token")\n       user = await client.get_user(7782553)\n       client.close()\n\n\n   if __name__ == "__main__":\n       asyncio.run(main())\n\n\nAPI v2 Example\n--------------\n\n.. code:: py\n\n    import aiosu\n    import asyncio\n    import datetime\n\n\n    async def main():\n        token = aiosu.classes.OAuthToken.parse_obj(json_token_from_api)\n\n        # or\n\n        token = aiosu.classes.OAuthToken(\n            access_token="access token",\n            refresh_token="refresh token",\n            expires_on=datetime.datetime.utcnow()\n            + datetime.timedelta(days=1),  # can also be string\n        )\n\n        # async with syntax\n        async with aiosu.v2.Client(\n            client_secret="secret", client_id=1000, token=token\n        ) as client:\n            user = await client.get_me()\n\n        # regular syntax\n        client = aiosu.v2.Client(client_secret="secret", client_id=1000, token=token)\n        user = await client.get_me()\n        client.close()\n\n\n    if __name__ == "__main__":\n        asyncio.run(main())\n\n\nYou can find more examples in the examples directory.\n\n\nContributing\n------------\n\nPlease read the `CONTRIBUTING.rst <.github/CONTRIBUTING.rst>`__ to learn how to contribute to aiosu!\n\n\nAcknowledgments\n---------------\n\n-  `discord.py <https://github.com/Rapptz/discord.py>`__\n   for README formatting\n-  `osu!Akatsuki <https://github.com/osuAkatsuki/performance-calculator>`__\n   for performance and accuracy utils\n\n\n.. |Python| image:: https://img.shields.io/pypi/pyversions/aiosu.svg\n    :target: https://pypi.python.org/pypi/aiosu\n    :alt: Python version info\n.. |pypi| image:: https://img.shields.io/pypi/v/aiosu.svg\n    :target: https://pypi.python.org/pypi/aiosu\n    :alt: PyPI version info\n.. |pre-commit.ci status| image:: https://results.pre-commit.ci/badge/github/NiceAesth/aiosu/master.svg\n    :target: https://results.pre-commit.ci/latest/github/NiceAesth/aiosu/master\n    :alt: pre-commit.ci status\n.. |pytest| image:: https://github.com/NiceAesth/aiosu/actions/workflows/pytest.yml/badge.svg\n    :target: https://github.com/NiceAesth/aiosu/actions/workflows/pytest.yml\n    :alt: pytest Status\n.. |mypy| image:: https://github.com/NiceAesth/aiosu/actions/workflows/mypy.yml/badge.svg\n    :target: https://github.com/NiceAesth/aiosu/actions/workflows/mypy.yml\n    :alt: mypy Status\n.. |rtd| image:: https://readthedocs.org/projects/aiosu/badge/?version=latest\n    :target: https://aiosu.readthedocs.io/en/latest/?badge=latest\n    :alt: Documentation Status\n',
    'author': 'Nice Aesthetics',
    'author_email': 'nice@aesth.dev',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/NiceAesth/aiosu',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



"""
This module handles API requests for API v1.

You can read more about it here: https://github.com/ppy/osu-api/wiki
"""
from __future__ import annotations

import functools
from typing import TYPE_CHECKING

import aiohttp
import orjson
from aiolimiter import AsyncLimiter

from .. import helpers
from ..classes import APIException
from ..classes import Beatmap
from ..classes import Beatmapset
from ..classes import Gamemode
from ..classes import Mods
from ..classes import Score
from ..classes import User
from ..classes import UserQueryType
from ..classes.legacy import Match
from ..classes.legacy import Replay

if TYPE_CHECKING:
    from types import TracebackType
    from typing import Any
    from typing import Callable
    from typing import Optional
    from typing import Type
    from typing import Union


def _beatmap_score_conv(data: Any, mode: Gamemode, beatmap_id: int) -> Score:
    data["beatmap_id"] = beatmap_id
    return Score._from_api_v1(data, mode)


def rate_limited(func: Callable) -> Callable:
    """
    A decorator that enforces rate limiting, to be used as:
    @rate_limited
    """

    @functools.wraps(func)
    async def _rate_limited(self: Client, *args: Any, **kwargs: Any) -> Any:
        if self._session is None:
            self._session = aiohttp.ClientSession()

        async with self._limiter:
            return await func(self, *args, **kwargs)

    return _rate_limited


class Client:
    r"""osu! API v1 Client

    :param token: The API key
    :type token: str
    :param \**kwargs:
        See below

    :Keyword Arguments:
        * *base_url* (``str``) --
            Optional, base API URL, defaults to \"https://osu.ppy.sh/api\"
        * *limiter* (``aiolimiter.AsyncLimiter``) --
            Optional, custom AsyncLimiter, defaults to AsyncLimiter(1200, 60)
    """

    def __init__(self, token: str, **kwargs: Any) -> None:
        self.token: str = token
        self.base_url: str = kwargs.pop("base_url", "https://osu.ppy.sh/api")
        self._limiter: AsyncLimiter = kwargs.pop("limiter", AsyncLimiter(1200, 60))
        self._session: aiohttp.ClientSession = None  # type: ignore

    async def __aenter__(self) -> Client:
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        await self.close()

    @rate_limited
    async def get_user(self, user_query: Union[str, int], **kwargs: Any) -> User:
        r"""Gets a user by a query.

        :param user_query: Username or ID to search by
        :type user_query: Union[str, int]
        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *mode* (``aiosu.classes.gamemode.Gamemode``) --
                Optional, gamemode to search for, defaults to standard
            * *qtype* (``str``) --
                Optional, \"string\" or \"id\". Type of the user_query
            * *event_days* (``aiosu.classes.gamemode.Gamemode``) --
                Optional, max number of days since last event, Min: 1, Max: 31, defaults to 1

        :raises APIException: Contains status code and error message
        :return: Requested user
        :rtype: list[aiosu.classes.user.User]
        """
        url = f"{self.base_url}/get_user"
        params = {
            "k": self.token,
            "u": user_query,
            "event_days": kwargs.pop("event_days", 1),
        }
        params["m"] = int(Gamemode(kwargs.pop("mode", 0)))  # type: ignore
        if "qtype" in kwargs:
            qtype = UserQueryType(kwargs.pop("qtype"))  # type: ignore
            params["type"] = qtype.old_api_name
        async with self._session.get(url, params=params) as resp:
            body = await resp.read()
            json = orjson.loads(body)
            if resp.status != 200:
                raise APIException(resp.status, json.get("error", ""))
            if not json:
                raise APIException(404, "User not found")
            return User._from_api_v1(json[0])

    @rate_limited
    async def __get_type_scores(
        self, user_query: Union[str, int], request_type: str, **kwargs: Any
    ) -> list[Score]:
        r"""INTERNAL: Get a user's scores by type

        :param user_query: Username or ID to search by
        :type user_query: Union[str, int]
        :param request_type: "recent" or "best"
        :type request_type: str
        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *mode* (``aiosu.classes.gamemode.Gamemode``) --
                Optional, gamemode to search for, defaults to standard
            * *limit* (``int``) --
                Optional, number of scores to get, defaults to 10
            * *qtype* (``str``) --
                Optional, \"string\" or \"id\". Type of the user_query

        :raises ValueError: If request_type is invalid
        :raises APIException: Contains status code and error message
        :return: List of requested scores
        :rtype: list[aiosu.classes.score.Score]
        """
        if request_type not in ("recent", "best"):
            raise ValueError(
                'Invalid request_type specified. Valid options are: "best", "recent"',
            )
        url = f"{self.base_url}/get_user_{request_type}"
        params = {
            "k": self.token,
            "u": user_query,
            "limit": kwargs.pop("limit", 10),
        }
        mode = Gamemode(kwargs.pop("mode", 0))  # type: ignore
        params["m"] = int(mode)
        if "qtype" in kwargs:
            qtype = UserQueryType(kwargs.pop("qtype"))  # type: ignore
            params["type"] = qtype.old_api_name
        async with self._session.get(url, params=params) as resp:
            body = await resp.read()
            json = orjson.loads(body)
            if resp.status != 200:
                raise APIException(resp.status, json.get("error", ""))
            score_conv = lambda x: Score._from_api_v1(x, mode)
            return helpers.from_list(score_conv, json)

    async def get_user_recents(
        self, user_query: Union[str, int], **kwargs: Any
    ) -> list[Score]:
        r"""Get a user's recent scores.

        :param user_query: Username or ID to search by
        :type user_query: Union[str, int]
        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *mode* (``aiosu.classes.gamemode.Gamemode``) --
                Optional, gamemode to search for, defaults to standard
            * *limit* (``int``) --
                Optional, number of scores to get, Min: 1, Max: 50, defaults to 50
            * *qtype* (``str``) --
                Optional, \"string\" or \"id\". Type of the user_query

        :raises ValueError: If limit is not between 1 and 50
        :raises APIException: Contains status code and error message
        :return: List of requested scores
        :rtype: list[aiosu.classes.score.Score]
        """
        limit = kwargs.pop("limit", 50)
        if not 1 <= limit <= 50:
            raise ValueError("Invalid limit specified. Limit must be between 1 and 50")
        return await self.__get_type_scores(user_query, "recent", limit=limit, **kwargs)

    async def get_user_bests(
        self, user_query: Union[str, int], **kwargs: Any
    ) -> list[Score]:
        r"""Get a user's best scores.

        :param user_query: Username or ID to search by
        :type user_query: Union[str, int]
        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *mode* (``aiosu.classes.gamemode.Gamemode``) --
                Optional, gamemode to search for, defaults to standard
            * *limit* (``int``) --
                Optional, number of scores to get, Min: 1, Max: 100, defaults to 100
            * *qtype* (``str``) --
                Optional, \"string\" or \"id\". Type of the user_query

        :raises ValueError: If limit is not between 1 and 100
        :raises APIException: Contains status code and error message
        :return: List of requested scores
        :rtype: list[aiosu.classes.score.Score]
        """
        limit = kwargs.pop("limit", 100)
        if not 1 <= limit <= 100:
            raise ValueError("Invalid limit specified. Limit must be between 1 and 100")
        return await self.__get_type_scores(user_query, "best", limit=limit, **kwargs)

    @rate_limited
    async def get_beatmap(self, **kwargs: Any) -> list[Beatmapset]:
        r"""Get beatmap data.

        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *limit* (``int``) --
                Optional, number of scores to get, Min: 1, Max: 500, defaults to 500
            * *mode* (``aiosu.classes.gamemode.Gamemode``) --
                Optional, gamemode to search for, defaults to standard
            * *converts* (``bool``) --
                Optional, whether to return converts, defaults to False
            * *mods* (``aiosu.classes.mods.Mods``) --
                Optional, mods to apply to the result
            * *beatmap_id* (``int``) --
                Optional, The ID of the beatmap
            * *beatmapset_id* (``int``) --
                Optional, The ID of the beatmapset
            * *since* (``datetime.datetime``) --
                Optional, Return all beatmaps with a leaderboard since this date
            * *hash* (``str``) --
                Optional, The MD5 hash of the beatmap
            * *user_query* (``Union[str, int]``) --
                Optional, username or ID to search by
            * *qtype* (``str``) --
                Optional, \"string\" or \"id\". Type of the user_query

        :raises ValueError: If limit is not between 1 and 500
        :raises ValueError: If none of hash, since, user_query, beatmap_id or beatmapset_id specified.
        :raises APIException: Contains status code and error message
        :return: List of beatmapsets each containing one difficulty of the result
        :rtype: list[aiosu.classes.beatmap.Beatmapset]
        """
        if not 1 <= kwargs.get("limit", 500) <= 500:
            raise ValueError("Invalid limit specified. Limit must be between 1 and 500")
        url = f"{self.base_url}/get_beatmaps"
        params = {
            "k": self.token,
            "limit": kwargs.pop("limit", 500),
            "a": int(kwargs.pop("converts", False)),
        }
        params["m"] = int(Gamemode(kwargs.pop("mode", 0)))  # type: ignore
        if "mods" in kwargs:
            mods = Mods(kwargs.pop("mods"))
            params["mode"] = str(mods)
        if "beatmap_id" in kwargs:
            params["b"] = kwargs.pop("beatmap_id")
        elif "beatmapset_id" in kwargs:
            params["s"] = kwargs.pop("beatmapset_id")
        elif "user_query" in kwargs:
            params["u"] = kwargs.pop("user_query")
            if "qtype" in kwargs:
                qtype = UserQueryType(kwargs.pop("qtype"))  # type: ignore
                params["type"] = qtype.old_api_name
        elif "since" in kwargs:
            params["since"] = kwargs.pop("since")
        elif "hash" in kwargs:
            params["h"] = kwargs.pop("hash")
        else:
            raise ValueError(
                "Either hash, since, user_query, beatmap_id or beatmapset_id must be specified.",
            )
        async with self._session.get(url, params=params) as resp:
            body = await resp.read()
            json = orjson.loads(body)
            if resp.status != 200:
                raise APIException(resp.status, json.get("error", ""))
            return helpers.from_list(Beatmapset._from_api_v1, json)

    @rate_limited
    async def get_beatmap_scores(self, beatmap_id: int, **kwargs: Any) -> list[Score]:
        r"""Get a user's best scores.

        :param beatmap_id: The ID of the beatmap
        :type beatmap_id: int
        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *mode* (``aiosu.classes.gamemode.Gamemode``) --
                Optional, gamemode to search for, defaults to standard
            * *mods* (``aiosu.classes.mods.Mods``) --
                Optional, mods to search for
            * *limit* (``int``) --
                Optional, number of scores to get, Min: 1, Max: 100, defaults to 100
            * *user_query* (``Union[str, int]``) --
                Optional, username or ID to search by
            * *qtype* (``str``) --
                Optional, \"string\" or \"id\". Type of the user_query

        :raises ValueError: If limit is not between 1 and 100
        :raises APIException: Contains status code and error message
        :return: List of requested scores
        :rtype: list[aiosu.classes.score.Score]
        """
        if not 1 <= kwargs.get("limit", 100) <= 100:
            raise ValueError("Invalid limit specified. Limit must be between 1 and 100")
        url = f"{self.base_url}/get_scores"
        params = {
            "k": self.token,
            "b": beatmap_id,
            "limit": kwargs.pop("limit", 50),
        }
        mode = Gamemode(kwargs.pop("mode", 0))  # type: ignore
        params["m"] = int(mode)
        if "user_query" in kwargs:
            params["u"] = kwargs.pop("user_query")
            if "qtype" in kwargs:
                qtype = UserQueryType(kwargs.pop("qtype"))  # type: ignore
                params["type"] = qtype.old_api_name
        if "mods" in kwargs:
            mods = Mods(kwargs.pop("mods"))
            params["mode"] = str(mods)
        async with self._session.get(url, params=params) as resp:
            body = await resp.read()
            json = orjson.loads(body)
            if resp.status != 200:
                raise APIException(resp.status, json.get("error", ""))
            score_conv = lambda x: _beatmap_score_conv(x, mode, beatmap_id)
            return helpers.from_list(score_conv, json)

    @rate_limited
    async def get_match(self, match_id: int) -> Match:
        r"""Gets a multiplayer match. (WIP, currently returns raw JSON)

        :param match_id: The ID of the match
        :type match_id: int
        :raises APIException: Contains status code and error message
        :return: The requested multiplayer match
        :rtype: aiosu.classes.legacy.match.Match
        """
        url = f"{self.base_url}/get_match"
        params = {
            "k": self.token,
            "mp": match_id,
        }
        async with self._session.get(url, params=params) as resp:
            body = await resp.read()
            json = orjson.loads(body)
            if resp.status != 200:
                raise APIException(resp.status, json.get("error", ""))
            return Match.parse_obj(json)

    @rate_limited
    async def get_replay(self, **kwargs: Any) -> Replay:
        r"""Gets data for a replay.

        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *mode* (``aiosu.classes.gamemode.Gamemode``) --
                Optional, gamemode to search for, defaults to standard
            * *mods* (``aiosu.classes.mods.Mods``) --
                Optional, mods to search for
            * *score_id* (``int``) --
                Optional, the ID of the score
            * *beatmap_id* (``int``) --
                Optional, the ID of the beatmap, specified together with user_query
            * *user_query* (``Union[str, int]``) --
                Optional, username or ID to search by, specified together with beatmap_id
            * *qtype* (``str``) --
                Optional, \"string\" or \"id\". Type of the user_query

        :raises ValueError: If neither score_id nor beatmap_id + user_id specified
        :raises APIException: Contains status code and error message
        :return: The data for the requested replay
        :rtype: aiosu.classes.legacy.replay.Replay
        """
        url = f"{self.base_url}/get_replay"
        params = {"k": self.token}
        params["m"] = int(Gamemode(kwargs.pop("mode", 0)))  # type: ignore
        if "score_id" in kwargs:
            params["s"] = kwargs.pop("score_id")
        elif "beatmap_id" in kwargs and "user_query" in kwargs:
            params["b"] = kwargs.pop("beatmap_id")
            params["u"] = kwargs.pop("user_query")
            if "qtype" in kwargs:
                qtype = UserQueryType(kwargs.pop("qtype"))  # type: ignore
                params["type"] = qtype.old_api_name
        else:
            raise ValueError(
                "Either score_id or beatmap_id + user_id must be specified.",
            )
        if "mods" in kwargs:
            mods = Mods(kwargs.pop("mods"))
            params["mode"] = str(mods)
        async with self._session.get(url, params=params) as resp:
            body = await resp.read()
            json = orjson.loads(body)
            if resp.status != 200:
                raise APIException(resp.status, json.get("error", ""))
            return Replay.parse_obj(json)

    async def close(self) -> None:
        """Closes the client session."""
        if self._session:
            await self._session.close()



"""
This module contains models for Score objects.
"""
from __future__ import annotations

import datetime
from typing import Optional
from typing import TYPE_CHECKING

from ..utils.accuracy import CatchAccuracyCalculator
from ..utils.accuracy import ManiaAccuracyCalculator
from ..utils.accuracy import OsuAccuracyCalculator
from ..utils.accuracy import TaikoAccuracyCalculator
from .beatmap import Beatmap
from .beatmap import Beatmapset
from .gamemode import Gamemode
from .models import BaseModel
from .mods import Mods
from .user import User

if TYPE_CHECKING:
    from typing import Any
    from .. import v1

accuracy_calculators = {
    "osu": OsuAccuracyCalculator(),
    "mania": ManiaAccuracyCalculator(),
    "taiko": TaikoAccuracyCalculator(),
    "fruits": CatchAccuracyCalculator(),
}


class ScoreWeight(BaseModel):
    percentage: float
    pp: float


class ScoreStatistics(BaseModel):
    count_50: int
    count_100: int
    count_300: int
    count_geki: int
    count_katu: int
    count_miss: int

    @classmethod
    def _from_api_v1(cls, data: Any) -> ScoreStatistics:
        return cls.parse_obj(
            {
                "count_50": data["count50"],
                "count_100": data["count100"],
                "count_300": data["count300"],
                "count_geki": data["countgeki"],
                "count_katu": data["countkatu"],
                "count_miss": data["countmiss"],
            },
        )


class Score(BaseModel):
    user_id: int
    accuracy: float
    mods: Mods
    score: int
    max_combo: int
    passed: bool
    perfect: bool
    statistics: ScoreStatistics
    rank: str
    created_at: datetime.datetime
    mode: Gamemode
    replay: bool
    id: Optional[int] = None
    """Always present except for API v1 recent scores."""
    pp: Optional[float] = 0
    best_id: Optional[int] = None
    beatmap: Optional[Beatmap] = None
    beatmapset: Optional[Beatmapset] = None
    weight: Optional[ScoreWeight] = None
    user: Optional[User] = None
    rank_global: Optional[int] = None
    rank_country: Optional[int] = None
    beatmap_id: Optional[int] = None
    """Only present on API v1"""

    @property
    def completion(self) -> float:  # Should probably move to utils
        """Beatmap completion.

        :raises ValueError: If beatmap is None
        :raises ValueError: If mode is unknown
        :return: Beatmap completion of a score (%). 100% for passes
        :rtype: float
        """
        if not self.beatmap:
            raise ValueError("Beatmap object is not set.")

        if self.mode == Gamemode.STANDARD:
            return (
                (
                    self.statistics.count_300
                    + self.statistics.count_100
                    + self.statistics.count_50
                    + self.statistics.count_miss
                )
                / self.beatmap.count_objects
            ) * 100
        if self.mode == Gamemode.TAIKO:
            return (
                (
                    self.statistics.count_300
                    + self.statistics.count_100
                    + self.statistics.count_miss
                )
                / self.beatmap.count_objects
            ) * 100
        if self.mode == Gamemode.CTB:
            return (
                (
                    self.statistics.count_300
                    + self.statistics.count_100
                    + +self.statistics.count_miss
                )
                / self.beatmap.count_objects
            ) * 100
        if self.mode == Gamemode.STANDARD:
            return (
                (
                    self.statistics.count_300
                    + self.statistics.count_100
                    + self.statistics.count_50
                    + self.statistics.count_miss
                    + self.statistics.count_geki
                    + self.statistics.count_katu
                )
                / self.beatmap.count_objects
            ) * 100
        raise ValueError("Unknown mode specified.")

    @property
    def score_url(self) -> Optional[str]:
        # score.id has undefined behaviour, best_id is the one you should use as it returns None if the URL does not exist
        """Link to the score.

        :return: Link to the score on the osu! website
        :rtype: Optional[str]
        """
        return (
            f"https://osu.ppy.sh/scores/{self.mode.name_api}/{self.best_id}"
            if self.best_id
            else None
        )

    @property
    def replay_url(self) -> Optional[str]:
        # score.id has undefined behaviour, best_id is the one you should use as it returns None if the URL does not exist
        """Link to the replay.

        :return: Link to download the replay on the osu! website
        :rtype: Optional[str]
        """
        return (
            f"https://osu.ppy.sh/scores/{self.mode.name_api}/{self.best_id}/download"
            if self.best_id
            else None
        )

    async def request_beatmap(self, client: v1.Client) -> None:
        """For v1 Scores: requests the beatmap from the API and sets it.

        :param client: An API v1 Client
        :type client: aiosu.v1.client.Client
        """
        if self.beatmap_id is None:
            raise ValueError("Score has unknown beatmap ID")
        if self.beatmap is None and self.beatmapset is None:
            sets = await client.get_beatmap(
                mode=self.mode,
                beatmap_id=self.beatmap_id,
            )
            self.beatmapset = sets[0]
            self.beatmap = sets[0].beatmaps[0]

    @classmethod
    def _from_api_v1(
        cls,
        data: Any,
        mode: Gamemode,
    ) -> Score:
        statistics = ScoreStatistics._from_api_v1(data)
        score = cls.parse_obj(
            {
                "id": data["score_id"],
                "user_id": data["user_id"],
                "accuracy": 0.0,
                "mods": int(data["enabled_mods"]),
                "score": data["score"],
                "pp": data.get("pp", 0.0),
                "max_combo": data["maxcombo"],
                "passed": data["rank"] != "F",
                "perfect": data["perfect"],
                "statistics": statistics,
                "rank": data["rank"],
                "created_at": data["date"],
                "mode": mode,
                "beatmap_id": data.get("beatmap_id"),
                "replay": data.get("replay_available", False),
            },
        )
        score.accuracy = accuracy_calculators[str(mode)].calculate(score)
        return score
